@@ -6,6 +6,7 @@
 
 struct server;
 struct output;
+struct seat;
 
 struct lab_layer_surface {
 	struct wlr_scene_layer_surface_v1 *scene_layer_surface;
@@ -36,5 +37,7 @@ struct lab_layer_popup {
 void layers_init(struct server *server);
 
 void layers_arrange(struct output *output);
+void layer_try_set_focus(struct seat *seat,
+	struct wlr_layer_surface_v1 *layer_surface);
 
 #endif /* LABWC_LAYERS_H */
