@@ -14,26 +14,6 @@ struct wlr_compositor;
 struct wlr_output;
 struct wlr_output_layout;
 
-enum atom {
-	/* https://specifications.freedesktop.org/wm-spec/wm-spec-1.4.html#idm45649101374512 */
-	NET_WM_WINDOW_TYPE_DESKTOP = 0,
-	NET_WM_WINDOW_TYPE_DOCK,
-	NET_WM_WINDOW_TYPE_TOOLBAR,
-	NET_WM_WINDOW_TYPE_MENU,
-	NET_WM_WINDOW_TYPE_UTILITY,
-	NET_WM_WINDOW_TYPE_SPLASH,
-	NET_WM_WINDOW_TYPE_DIALOG,
-	NET_WM_WINDOW_TYPE_DROPDOWN_MENU,
-	NET_WM_WINDOW_TYPE_POPUP_MENU,
-	NET_WM_WINDOW_TYPE_TOOLTIP,
-	NET_WM_WINDOW_TYPE_NOTIFICATION,
-	NET_WM_WINDOW_TYPE_COMBO,
-	NET_WM_WINDOW_TYPE_DND,
-	NET_WM_WINDOW_TYPE_NORMAL,
-
-	ATOM_LEN
-};
-
 static const char * const atom_names[] = {
 	"_NET_WM_WINDOW_TYPE_DESKTOP",
 	"_NET_WM_WINDOW_TYPE_DOCK",
@@ -52,10 +32,10 @@ static const char * const atom_names[] = {
 };
 
 static_assert(
-	ARRAY_SIZE(atom_names) == ATOM_LEN,
+	ARRAY_SIZE(atom_names) == WINDOW_TYPE_LEN,
 	"Xwayland atoms out of sync");
 
-extern xcb_atom_t atoms[ATOM_LEN];
+extern xcb_atom_t atoms[WINDOW_TYPE_LEN];
 
 struct xwayland_unmanaged {
 	struct server *server;
@@ -105,9 +85,6 @@ void xwayland_adjust_stacking_order(struct server *server);
 
 struct wlr_xwayland_surface *xwayland_surface_from_view(struct view *view);
 
-bool xwayland_surface_contains_window_type(
-	struct wlr_xwayland_surface *surface, enum atom window_type);
-
 void xwayland_server_init(struct server *server,
 	struct wlr_compositor *compositor);
 void xwayland_server_finish(struct server *server);
