@@ -11,12 +11,14 @@
 #include "common/macros.h"
 #include "common/mem.h"
 #include "common/scene-helpers.h"
+#include "common/surface-helpers.h"
 #include "config/mousebind.h"
 #include "dnd.h"
 #include "idle.h"
 #include "input/gestures.h"
 #include "input/touch.h"
 #include "labwc.h"
+#include "layers.h"
 #include "menu/menu.h"
 #include "regions.h"
 #include "resistance.h"
@@ -223,16 +225,7 @@ process_cursor_move(struct server *server, uint32_t time)
 	resistance_move_apply(view, &dx, &dy);
 	view_move(view, dx, dy);
 
-	/* Region overlay */
-	if (!regions_should_snap(server)) {
-		return;
-	}
-	struct region *region = regions_from_cursor(server);
-	if (region) {
-		regions_show_overlay(view, &server->seat, region);
-	} else {
-		regions_hide_overlay(&server->seat);
-	}
+	overlay_update(&server->seat);
 }
 
 static void
@@ -306,6 +299,7 @@ cursor_update_image(struct seat *seat)
 		 */
 		if (seat->seat->pointer_state.focused_surface) {
 			seat->server_cursor = LAB_CURSOR_DEFAULT;
+			wlr_cursor_set_xcursor(seat->cursor, seat->xcursor_manager, "");
 			cursor_update_focus(seat->server);
 		}
 		return;
@@ -527,7 +521,8 @@ process_cursor_motion(struct server *server, uint32_t time)
 		dnd_icons_move(seat, seat->cursor->x, seat->cursor->y);
 	}
 
-	if ((ctx.view || ctx.surface) && rc.focus_follow_mouse) {
+	if ((ctx.view || ctx.surface) && rc.focus_follow_mouse
+			&& !server->osd_state.cycle_view) {
 		desktop_focus_view_or_surface(seat, ctx.view, ctx.surface,
 			rc.raise_on_focus);
 	}
@@ -804,6 +799,10 @@ static bool
 handle_release_mousebinding(struct server *server,
 		struct cursor_context *ctx, uint32_t button)
 {
+	if (server->osd_state.cycle_view) {
+		return false;
+	}
+
 	struct mousebind *mousebind;
 	bool consumed_by_frame_context = false;
 
@@ -853,8 +852,10 @@ handle_release_mousebinding(struct server *server,
 }
 
 static bool
-is_double_click(long double_click_speed, uint32_t button, struct view *view)
+is_double_click(long double_click_speed, uint32_t button,
+		struct cursor_context *ctx)
 {
+	static enum ssd_part_type last_type;
 	static uint32_t last_button;
 	static struct view *last_view;
 	static struct timespec last_click;
@@ -864,9 +865,11 @@ is_double_click(long double_click_speed, uint32_t button, struct view *view)
 	long ms = (now.tv_sec - last_click.tv_sec) * 1000 +
 		(now.tv_nsec - last_click.tv_nsec) / 1000000;
 	last_click = now;
-	if (last_button != button || last_view != view) {
+	if (last_button != button || last_view != ctx->view
+			|| last_type != ctx->type) {
 		last_button = button;
-		last_view = view;
+		last_view = ctx->view;
+		last_type = ctx->type;
 		return false;
 	}
 	if (ms < double_click_speed && ms >= 0) {
@@ -876,6 +879,7 @@ is_double_click(long double_click_speed, uint32_t button, struct view *view)
 		 */
 		last_button = 0;
 		last_view = NULL;
+		last_type = 0;
 		return true;
 	}
 	return false;
@@ -885,8 +889,12 @@ static bool
 handle_press_mousebinding(struct server *server, struct cursor_context *ctx,
 		uint32_t button, uint32_t resize_edges)
 {
+	if (server->osd_state.cycle_view) {
+		return false;
+	}
+
 	struct mousebind *mousebind;
-	bool double_click = is_double_click(rc.doubleclick_time, button, ctx->view);
+	bool double_click = is_double_click(rc.doubleclick_time, button, ctx);
 	bool consumed_by_frame_context = false;
 
 	uint32_t modifiers = wlr_keyboard_get_modifiers(
@@ -931,8 +939,7 @@ handle_press_mousebinding(struct server *server, struct cursor_context *ctx,
 	return consumed_by_frame_context;
 }
 
-/* Set in cursor_button_press(), used in cursor_button_release() */
-static bool close_menu;
+static uint32_t press_msec;
 
 static void
 cursor_button_press(struct seat *seat, uint32_t button,
@@ -941,6 +948,9 @@ cursor_button_press(struct seat *seat, uint32_t button,
 	struct server *server = seat->server;
 	struct cursor_context ctx = get_cursor_context(server);
 
+	/* Used on next button release to check if it can close menu or select menu item */
+	press_msec = time_msec;
+
 	/* Determine closest resize edges in case action is Resize */
 	uint32_t resize_edges = cursor_get_resize_edges(seat->cursor, &ctx);
 
@@ -952,10 +962,10 @@ cursor_button_press(struct seat *seat, uint32_t button,
 
 	if (server->input_mode == LAB_INPUT_STATE_MENU) {
 		/*
-		 * We close the menu on RELEASE to not leak a stray releases and
-		 * to be consistent with Openbox
+		 * If menu was already opened on press, set a very small value
+		 * so subsequent release always closes menu or selects menu item.
 		 */
-		close_menu = true;
+		press_msec = 0;
 		return;
 	}
 
@@ -965,10 +975,18 @@ cursor_button_press(struct seat *seat, uint32_t button,
 	 * the Focus action (used for normal views) does not work.
 	 */
 	if (ctx.type == LAB_SSD_LAYER_SURFACE) {
+		wlr_log(WLR_DEBUG, "press on layer-surface");
 		struct wlr_layer_surface_v1 *layer =
 			wlr_layer_surface_v1_try_from_wlr_surface(ctx.surface);
 		if (layer && layer->current.keyboard_interactive) {
-			seat_set_focus_layer(seat, layer);
+			layer_try_set_focus(seat, layer);
+		}
+	} else if (ctx.type == LAB_SSD_LAYER_SUBSURFACE) {
+		wlr_log(WLR_DEBUG, "press on layer-subsurface");
+		struct wlr_layer_surface_v1 *layer =
+			subsurface_parent_layer(ctx.surface);
+		if (layer && layer->current.keyboard_interactive) {
+			layer_try_set_focus(seat, layer);
 		}
 #ifdef HAVE_XWAYLAND
 	} else if (ctx.type == LAB_SSD_UNMANAGED) {
@@ -977,12 +995,6 @@ cursor_button_press(struct seat *seat, uint32_t button,
 #endif
 	}
 
-	/*
-	 * TODO: We may need to handle press on layer-shell subsurfaces here,
-	 * but need to check keyboard interactivity before focusing them
-	 * otherwise we break waybar. See issue #1131
-	 */
-
 	if (ctx.type != LAB_SSD_CLIENT && ctx.type != LAB_SSD_LAYER_SUBSURFACE
 			&& wlr_seat_pointer_has_grab(seat->seat)) {
 		/*
@@ -1019,7 +1031,8 @@ cursor_button_release(struct seat *seat, uint32_t button,
 	seat_reset_pressed(seat);
 
 	if (server->input_mode == LAB_INPUT_STATE_MENU) {
-		if (close_menu) {
+		/* TODO: take into account overflow of time_msec */
+		if (time_msec - press_msec > rc.menu_ignore_button_release_period) {
 			if (ctx.type == LAB_SSD_MENU) {
 				menu_call_selected_actions(server);
 			} else {
@@ -1027,20 +1040,20 @@ cursor_button_release(struct seat *seat, uint32_t button,
 				cursor_update_common(server, &ctx, time_msec,
 					/*cursor_has_moved*/ false);
 			}
-			close_menu = false;
 		}
 		return;
 	}
 
 	if (server->input_mode != LAB_INPUT_STATE_PASSTHROUGH) {
-		/* Exit interactive move/resize mode */
-		interactive_finish(server->grabbed_view);
-
 		if (pressed_surface) {
 			/* Ensure CSD clients see the release event */
 			wlr_seat_pointer_notify_button(seat->seat, time_msec,
 				button, button_state);
 		}
+
+		/* Exit interactive move/resize mode */
+		interactive_finish(server->grabbed_view);
+
 		return;
 	}
 
@@ -1058,7 +1071,7 @@ cursor_button_release(struct seat *seat, uint32_t button,
 	bool consumed_by_frame_context =
 		handle_release_mousebinding(server, &ctx, button);
 
-	if (ctx.surface && !consumed_by_frame_context) {
+	if (!consumed_by_frame_context) {
 		/* Notify client with pointer focus of button release */
 		wlr_seat_pointer_notify_button(seat->seat, time_msec,
 			button, button_state);
@@ -1090,6 +1103,8 @@ void
 cursor_emulate_move_absolute(struct seat *seat, struct wlr_input_device *device,
 		double x, double y, uint32_t time_msec)
 {
+	idle_manager_notify_activity(seat->seat);
+
 	double lx, ly;
 	wlr_cursor_absolute_to_layout_coords(seat->cursor,
 		device, x, y, &lx, &ly);
@@ -1249,13 +1264,16 @@ cursor_frame(struct wl_listener *listener, void *data)
 	wlr_seat_pointer_notify_frame(seat->seat);
 }
 
-void
-cursor_init(struct seat *seat)
+static void
+cursor_load(struct seat *seat)
 {
 	const char *xcursor_theme = getenv("XCURSOR_THEME");
 	const char *xcursor_size = getenv("XCURSOR_SIZE");
 	uint32_t size = xcursor_size ? atoi(xcursor_size) : 24;
 
+	if (seat->xcursor_manager) {
+		wlr_xcursor_manager_destroy(seat->xcursor_manager);
+	}
 	seat->xcursor_manager = wlr_xcursor_manager_create(xcursor_theme, size);
 	wlr_xcursor_manager_load(seat->xcursor_manager, 1);
 
@@ -1290,6 +1308,19 @@ cursor_init(struct seat *seat)
 			"Cursor theme is missing cursor names, using fallback");
 		cursor_names = cursors_x11;
 	}
+}
+
+void
+cursor_reload(struct seat *seat)
+{
+	cursor_load(seat);
+	cursor_update_image(seat);
+}
+
+void
+cursor_init(struct seat *seat)
+{
+	cursor_load(seat);
 
 	/* Set the initial cursor image so the cursor is visible right away */
 	cursor_set(seat, LAB_CURSOR_DEFAULT);
