@@ -9,6 +9,7 @@
 #include <wlr/types/wlr_touch.h>
 #include <wlr/util/log.h>
 #include "common/mem.h"
+#include "input/ime.h"
 #include "input/tablet.h"
 #include "input/tablet_pad.h"
 #include "input/input.h"
@@ -30,6 +31,19 @@ input_device_destroy(struct wl_listener *listener, void *data)
 		wl_list_remove(&keyboard->key.link);
 		wl_list_remove(&keyboard->modifier.link);
 		keyboard_cancel_keybind_repeat(keyboard);
+		/*
+		 * If the active keyboard on the seat is destroyed, fall back
+		 * to the keyboard from keyboard group so we can respond to
+		 * wl_seat.get_keyboard requests with wl_keyboard.keymap event.
+		 * This prevents Chromium from crashing when started just after
+		 * the active keyboard is destroyed.
+		 */
+		struct wlr_keyboard *active_keyboard =
+			wlr_seat_get_keyboard(input->seat->seat);
+		if (!active_keyboard || active_keyboard == keyboard->wlr_keyboard) {
+			wlr_seat_set_keyboard(input->seat->seat,
+				&input->seat->keyboard_group->keyboard);
+		}
 	}
 	free(input);
 }
@@ -98,7 +112,7 @@ configure_libinput(struct wlr_input_device *wlr_input_device)
 	 *       that some libinput setting could not be applied.
 	 *
 	 * TODO: We are currently using int32_t with -1 as default
-	 *       to desribe the not-configured state. This is not
+	 *       to describe the not-configured state. This is not
 	 *       really optimal as we can't properly deal with
 	 *       enum values that are 0. After some discussion via
 	 *       IRC the best way forward seem to be to use a
@@ -237,6 +251,15 @@ configure_libinput(struct wlr_input_device *wlr_input_device)
 		wlr_log(WLR_INFO, "send events mode configured");
 		libinput_device_config_send_events_set_mode(libinput_dev, dc->send_events_mode);
 	}
+
+	/* Non-zero if the device can be calibrated, zero otherwise. */
+	if (libinput_device_config_calibration_has_matrix(libinput_dev) == 0
+			|| !dc->have_calibration_matrix) {
+		wlr_log(WLR_INFO, "calibration matrix not configured");
+	} else {
+		wlr_log(WLR_INFO, "calibration matrix configured");
+		libinput_device_config_calibration_set_matrix(libinput_dev, dc->calibration_matrix);
+	}
 }
 
 static struct wlr_output *
@@ -296,6 +319,11 @@ new_keyboard(struct seat *seat, struct wlr_input_device *device, bool virtual)
 	keyboard->wlr_keyboard = kb;
 	keyboard->is_virtual = virtual;
 
+	if (!seat->keyboard_group->keyboard.keymap) {
+		wlr_log(WLR_ERROR, "cannot set keymap");
+		exit(EXIT_FAILURE);
+	}
+
 	wlr_keyboard_set_keymap(kb, seat->keyboard_group->keyboard.keymap);
 
 	/*
@@ -474,20 +502,11 @@ focus_change_notify(struct wl_listener *listener, void *data)
 	struct wlr_surface *surface = event->new_surface;
 	struct view *view = surface ? view_from_wlr_surface(surface) : NULL;
 
-	/* Prevent focus switch to layershell client from updating view state */
-	if (surface && wlr_layer_surface_v1_try_from_wlr_surface(surface)) {
-		return;
-	}
-
 	/*
-	 * If an xwayland-unmanaged surface was focused belonging to the
-	 * same application as the focused view, allow the view to remain
-	 * active. This fixes an issue with menus immediately closing in
-	 * some X11 apps (try LibreOffice with SAL_USE_VCLPLUGIN=gen).
+	 * Prevent focus switch to non-view surface (e.g. layer-shell
+	 * or xwayland-unmanaged) from updating view state
 	 */
-	if (!view && server->active_view && event->new_surface
-			&& view_is_related(server->active_view,
-				event->new_surface)) {
+	if (surface && !view) {
 		return;
 	}
 
@@ -536,6 +555,9 @@ seat_init(struct server *server)
 		&seat->virtual_keyboard_new);
 	seat->virtual_keyboard_new.notify = new_virtual_keyboard;
 
+	seat->input_method_relay = input_method_relay_create(seat);
+
+	seat->xcursor_manager = NULL;
 	seat->cursor = wlr_cursor_create();
 	if (!seat->cursor) {
 		wlr_log(WLR_ERROR, "unable to create cursor");
@@ -559,6 +581,7 @@ seat_finish(struct server *server)
 	}
 
 	input_handlers_finish(seat);
+	input_method_relay_finish(seat->input_method_relay);
 }
 
 static void
@@ -577,6 +600,8 @@ seat_reconfigure(struct server *server)
 {
 	struct seat *seat = &server->seat;
 	struct input *input;
+	cursor_reload(seat);
+	overlay_reconfigure(seat);
 	wl_list_for_each(input, &seat->inputs, link) {
 		switch (input->wlr_input_device->type) {
 		case WLR_INPUT_DEVICE_KEYBOARD:
@@ -614,6 +639,7 @@ seat_focus(struct seat *seat, struct wlr_surface *surface, bool is_lock_surface)
 
 	if (!surface) {
 		wlr_seat_keyboard_notify_clear_focus(seat->seat);
+		input_method_relay_set_focus(seat->input_method_relay, NULL);
 		return;
 	}
 
@@ -636,6 +662,8 @@ seat_focus(struct seat *seat, struct wlr_surface *surface, bool is_lock_surface)
 	wlr_seat_keyboard_notify_enter(seat->seat, surface,
 		pressed_sent_keycodes, nr_pressed_sent_keycodes, &kb->modifiers);
 
+	input_method_relay_set_focus(seat->input_method_relay, surface);
+
 	struct wlr_pointer_constraint_v1 *constraint =
 		wlr_pointer_constraints_v1_constraint_for_surface(server->constraints,
 			surface, seat->seat);
@@ -668,9 +696,7 @@ seat_set_focus_layer(struct seat *seat, struct wlr_layer_surface_v1 *layer)
 		return;
 	}
 	seat_focus(seat, layer->surface, /*is_lock_surface*/ false);
-	if (layer->current.layer >= ZWLR_LAYER_SHELL_V1_LAYER_TOP) {
-		seat->focused_layer = layer;
-	}
+	seat->focused_layer = layer;
 }
 
 static void
