@@ -2,6 +2,7 @@
 #define _POSIX_C_SOURCE 200809L
 #include "config.h"
 #include <signal.h>
+#include <string.h>
 #include <sys/wait.h>
 #include <wlr/backend/headless.h>
 #include <wlr/backend/multi.h>
@@ -17,6 +18,7 @@
 #include <wlr/types/wlr_viewporter.h>
 #if HAVE_XWAYLAND
 #include <wlr/xwayland.h>
+#include "xwayland-shell-v1-protocol.h"
 #endif
 #include "drm-lease-v1-protocol.h"
 #include "config/rcxml.h"
@@ -26,6 +28,7 @@
 #include "labwc.h"
 #include "layers.h"
 #include "menu/menu.h"
+#include "output-virtual.h"
 #include "regions.h"
 #include "resize_indicator.h"
 #include "theme.h"
@@ -40,34 +43,37 @@ static struct wlr_compositor *compositor;
 static struct wl_event_source *sighup_source;
 static struct wl_event_source *sigint_source;
 static struct wl_event_source *sigterm_source;
-
-static struct server *g_server;
+static struct wl_event_source *sigchld_source;
 
 static void
-reload_config_and_theme(void)
+reload_config_and_theme(struct server *server)
 {
 	rcxml_finish();
 	rcxml_read(rc.config_file);
-	theme_finish(g_server->theme);
-	theme_init(g_server->theme, rc.theme_name);
+	theme_finish(server->theme);
+	theme_init(server->theme, server, rc.theme_name);
 
 	struct view *view;
-	wl_list_for_each(view, &g_server->views, link) {
+	wl_list_for_each(view, &server->views, link) {
 		view_reload_ssd(view);
 	}
 
-	menu_reconfigure(g_server);
-	seat_reconfigure(g_server);
-	regions_reconfigure(g_server);
-	resize_indicator_reconfigure(g_server);
+	menu_reconfigure(server);
+	seat_reconfigure(server);
+	regions_reconfigure(server);
+	resize_indicator_reconfigure(server);
 	kde_server_decoration_update_default();
+	workspaces_reconfigure(server);
 }
 
 static int
 handle_sighup(int signal, void *data)
 {
+	struct server *server = data;
+
 	session_environment_init();
-	reload_config_and_theme();
+	reload_config_and_theme(server);
+	output_virtual_update_fallback(server);
 	return 0;
 }
 
@@ -80,6 +86,68 @@ handle_sigterm(int signal, void *data)
 	return 0;
 }
 
+static int
+handle_sigchld(int signal, void *data)
+{
+	siginfo_t info;
+	info.si_pid = 0;
+	struct server *server = data;
+
+	/* First call waitid() with NOWAIT which doesn't consume the zombie */
+	if (waitid(P_ALL, /*id*/ 0, &info, WEXITED | WNOHANG | WNOWAIT) == -1) {
+		return 0;
+	}
+
+	if (info.si_pid == 0) {
+		/* No children in waitable state */
+		return 0;
+	}
+
+#if HAVE_XWAYLAND
+	/* Ensure that we do not break xwayland lazy initialization */
+	if (server->xwayland && server->xwayland->server
+			&& info.si_pid == server->xwayland->server->pid) {
+		return 0;
+	}
+#endif
+
+	/* And then do the actual (consuming) lookup again */
+	int ret = waitid(P_PID, info.si_pid, &info, WEXITED);
+	if (ret == -1) {
+		wlr_log(WLR_ERROR, "blocking waitid() for %ld failed: %d",
+			(long)info.si_pid, ret);
+		return 0;
+	}
+
+	switch (info.si_code) {
+	case CLD_EXITED:
+		wlr_log(info.si_status == 0 ? WLR_DEBUG : WLR_ERROR,
+			"spawned child %ld exited with %d",
+			(long)info.si_pid, info.si_status);
+		break;
+	case CLD_KILLED:
+	case CLD_DUMPED:
+		; /* works around "a label can only be part of a statement" */
+		const char *signame = strsignal(info.si_status);
+		wlr_log(WLR_ERROR,
+			"spawned child %ld terminated with signal %d (%s)",
+				(long)info.si_pid, info.si_status,
+				signame ? signame : "unknown");
+		break;
+	default:
+		wlr_log(WLR_ERROR,
+			"spawned child %ld terminated unexpectedly: %d"
+			" please report", (long)info.si_pid, info.si_code);
+	}
+
+	if (info.si_pid == server->primary_client_pid) {
+		wlr_log(WLR_INFO, "primary client %ld exited", (long)info.si_pid);
+		wl_display_terminate(server->wl_display);
+	}
+
+	return 0;
+}
+
 static void
 seat_inhibit_input(struct seat *seat,  struct wl_client *active_client)
 {
@@ -172,9 +240,11 @@ server_global_filter(const struct wl_client *client, const struct wl_global *glo
 	(void)iface; (void)server;
 
 #if HAVE_XWAYLAND
-	struct wl_client *xwayland_client =
-		server->xwayland ? server->xwayland->server->client : NULL;
-	if (xwayland_client && client == xwayland_client) {
+	struct wl_client *xwayland_client = (server->xwayland && server->xwayland->server)
+		? server->xwayland->server->client
+		: NULL;
+
+	if (client == xwayland_client) {
 		/*
 		 * Filter out wp_drm_lease_device_v1 for now as it is resulting in
 		 * issues with Xwayland applications lagging over time.
@@ -184,6 +254,9 @@ server_global_filter(const struct wl_client *client, const struct wl_global *glo
 		if (!strcmp(iface->name, wp_drm_lease_device_v1_interface.name)) {
 			return false;
 		}
+	} else if (!strcmp(iface->name, xwayland_shell_v1_interface.name)) {
+		/* Filter out the xwayland shell for usual clients */
+		return false;
 	}
 #endif
 
@@ -211,9 +284,19 @@ static const char helpful_seat_error_message[] =
 "If the above does not work, try running with `WLR_RENDERER=pixman labwc` in\n"
 "order to use the software rendering fallback\n";
 
+static void
+get_headless_backend(struct wlr_backend *backend, void *data)
+{
+	if (wlr_backend_is_headless(backend)) {
+		struct wlr_backend **headless = data;
+		*headless = backend;
+	}
+}
+
 void
 server_init(struct server *server)
 {
+	server->primary_client_pid = -1;
 	server->wl_display = wl_display_create();
 	if (!server->wl_display) {
 		wlr_log(WLR_ERROR, "cannot allocate a wayland display");
@@ -226,11 +309,13 @@ server_init(struct server *server)
 	struct wl_event_loop *event_loop = NULL;
 	event_loop = wl_display_get_event_loop(server->wl_display);
 	sighup_source = wl_event_loop_add_signal(
-		event_loop, SIGHUP, handle_sighup, NULL);
+		event_loop, SIGHUP, handle_sighup, server);
 	sigint_source = wl_event_loop_add_signal(
 		event_loop, SIGINT, handle_sigterm, server->wl_display);
 	sigterm_source = wl_event_loop_add_signal(
 		event_loop, SIGTERM, handle_sigterm, server->wl_display);
+	sigchld_source = wl_event_loop_add_signal(
+		event_loop, SIGCHLD, handle_sigchld, server);
 	server->wl_event_loop = event_loop;
 
 	/*
@@ -258,7 +343,16 @@ server_init(struct server *server)
 	}
 
 	/* Create headless backend to enable adding virtual outputs later on */
-	server->headless.backend = wlr_headless_backend_create(server->wl_display);
+	wlr_multi_for_each_backend(server->backend,
+		get_headless_backend, &server->headless.backend);
+
+	if (!server->headless.backend) {
+		wlr_log(WLR_DEBUG, "manually creating headless backend");
+		server->headless.backend = wlr_headless_backend_create(server->wl_display);
+	} else {
+		wlr_log(WLR_DEBUG, "headless backend already exists");
+	}
+
 	if (!server->headless.backend) {
 		wlr_log(WLR_ERROR, "unable to create headless backend");
 		exit(EXIT_FAILURE);
@@ -378,6 +472,10 @@ server_init(struct server *server)
 	 */
 	wlr_primary_selection_v1_device_manager_create(server->wl_display);
 
+	server->input_method_manager = wlr_input_method_manager_v2_create(
+		server->wl_display);
+	server->text_input_manager = wlr_text_input_manager_v3_create(
+		server->wl_display);
 	seat_init(server);
 	xdg_shell_init(server);
 	kde_server_decoration_init(server);
@@ -457,8 +555,6 @@ server_init(struct server *server)
 #if HAVE_XWAYLAND
 	xwayland_server_init(server, compositor);
 #endif
-	/* used when handling SIGHUP */
-	g_server = server;
 }
 
 void
@@ -480,6 +576,9 @@ server_start(struct server *server)
 		exit(EXIT_FAILURE);
 	}
 
+	/* Potentially set up the initial fallback output */
+	output_virtual_update_fallback(server);
+
 	if (setenv("WAYLAND_DISPLAY", socket, true) < 0) {
 		wlr_log_errno(WLR_ERROR, "unable to set WAYLAND_DISPLAY");
 	} else {
