@@ -17,6 +17,7 @@
 #include <string.h>
 #include <wlr/util/box.h>
 #include <wlr/util/log.h>
+#include <wlr/render/pixman.h>
 #include <strings.h>
 #include "common/macros.h"
 #include "common/dir.h"
@@ -24,9 +25,12 @@
 #include "common/graphic-helpers.h"
 #include "common/match.h"
 #include "common/mem.h"
+#include "common/parse-bool.h"
+#include "common/parse-double.h"
 #include "common/string-helpers.h"
 #include "config/rcxml.h"
 #include "button/button-png.h"
+#include "labwc.h"
 
 #if HAVE_RSVG
 #include "button/button-svg.h"
@@ -135,8 +139,8 @@ create_hover_fallback(struct theme *theme, const char *icon_name,
 		(width - icon_width) / 2, (height - icon_height) / 2);
 	cairo_paint(cairo);
 
-	/* Overlay (non-multiplied alpha) */
-	float overlay_color[4] = { 0.5f, 0.5f, 0.5f, 0.3f};
+	/* Overlay (pre-multiplied alpha) */
+	float overlay_color[4] = { 0.15f, 0.15f, 0.15f, 0.3f};
 	enum corner corner = corner_from_icon_name(icon_name);
 
 	if (corner == LAB_CORNER_UNKNOWN) {
@@ -167,7 +171,7 @@ create_hover_fallback(struct theme *theme, const char *icon_name,
 
 /*
  * We use the following button filename schema: "BUTTON [TOGGLED] [STATE]"
- * with the words separted by underscore, and the following meaning:
+ * with the words separated by underscore, and the following meaning:
  *   - BUTTON can be one of 'max', 'iconify', 'close', 'menu'
  *   - TOGGLED is either 'toggled' or nothing
  *   - STATE is 'hover' or nothing. In future, 'pressed' may be supported too.
@@ -404,11 +408,37 @@ parse_hexstr(const char *hex, float *rgba)
 	rgba[0] = (hex_to_dec(hex[1]) * 16 + hex_to_dec(hex[2])) / 255.0;
 	rgba[1] = (hex_to_dec(hex[3]) * 16 + hex_to_dec(hex[4])) / 255.0;
 	rgba[2] = (hex_to_dec(hex[5]) * 16 + hex_to_dec(hex[6])) / 255.0;
-	if (strlen(hex) > 7) {
-		rgba[3] = atoi(hex + 7) / 100.0;
-	} else {
-		rgba[3] = 1.0;
+	rgba[3] = 1.0;
+
+	size_t len = strlen(hex);
+	if (len >= 9 && hex[7] == ' ') {
+		/* Deprecated #aabbcc 100 alpha encoding to support openbox themes */
+		rgba[3] = atoi(hex + 8) / 100.0;
+		wlr_log(WLR_ERROR,
+			"The theme uses deprecated alpha notation %s, please convert to "
+			"#rrggbbaa to ensure your config works on newer labwc releases", hex);
+	} else if (len == 9) {
+		/* Inline alpha encoding like #aabbccff */
+		rgba[3] = (hex_to_dec(hex[7]) * 16 + hex_to_dec(hex[8])) / 255.0;
+	} else if (len > 7) {
+		/* More than just #aabbcc */
+		wlr_log(WLR_ERROR, "invalid alpha color encoding: '%s'", hex);
 	}
+
+	/* Pre-multiply everything as expected by wlr_scene */
+	rgba[0] *= rgba[3];
+	rgba[1] *= rgba[3];
+	rgba[2] *= rgba[3];
+}
+
+static void
+parse_hexstrs(const char *hexes, float colors[3][4])
+{
+	gchar **elements = g_strsplit(hexes, ",", -1);
+	for (int i = 0; elements[i] && i < 3; i++) {
+		parse_hexstr(elements[i], colors[i]);
+	}
+	g_strfreev(elements);
 }
 
 static enum lab_justification
@@ -436,7 +466,7 @@ parse_justification(const char *str)
  * theme_builtin() applies a theme that is similar to vanilla GTK
  */
 static void
-theme_builtin(struct theme *theme)
+theme_builtin(struct theme *theme, struct server *server)
 {
 	theme->border_width = 1;
 	theme->padding_height = 3;
@@ -473,6 +503,11 @@ theme_builtin(struct theme *theme)
 	parse_hexstr("#000000",
 		theme->window_inactive_button_close_unpressed_image_color);
 
+	theme->window_active_shadow_size = 60;
+	theme->window_inactive_shadow_size = 40;
+	parse_hexstr("#00000060", theme->window_active_shadow_color);
+	parse_hexstr("#00000040", theme->window_inactive_shadow_color);
+
 	parse_hexstr("#fcfbfa", theme->menu_items_bg_color);
 	parse_hexstr("#000000", theme->menu_items_text_color);
 	parse_hexstr("#e1dedb", theme->menu_items_active_bg_color);
@@ -490,11 +525,18 @@ theme_builtin(struct theme *theme)
 	parse_hexstr("#888888", theme->menu_separator_color);
 
 	theme->osd_window_switcher_width = 600;
+	theme->osd_window_switcher_width_is_percent = false;
 	theme->osd_window_switcher_padding = 4;
 	theme->osd_window_switcher_item_padding_x = 10;
 	theme->osd_window_switcher_item_padding_y = 1;
 	theme->osd_window_switcher_item_active_border_width = 2;
 
+	/* inherit settings in post_processing() if not set elsewhere */
+	theme->osd_window_switcher_preview_border_width = INT_MIN;
+	memset(theme->osd_window_switcher_preview_border_color, 0,
+		sizeof(theme->osd_window_switcher_preview_border_color));
+	theme->osd_window_switcher_preview_border_color[0][0] = FLT_MIN;
+
 	theme->osd_workspace_switcher_boxes_width = 20;
 	theme->osd_workspace_switcher_boxes_height = 20;
 
@@ -503,6 +545,32 @@ theme_builtin(struct theme *theme)
 	theme->osd_border_width = INT_MIN;
 	theme->osd_border_color[0] = FLT_MIN;
 	theme->osd_label_text_color[0] = FLT_MIN;
+
+	if (wlr_renderer_is_pixman(server->renderer)) {
+		/* Draw only outlined overlay by default to save CPU resource */
+		theme->snapping_overlay_region.bg_enabled = false;
+		theme->snapping_overlay_edge.bg_enabled = false;
+		theme->snapping_overlay_region.border_enabled = true;
+		theme->snapping_overlay_edge.border_enabled = true;
+	} else {
+		theme->snapping_overlay_region.bg_enabled = true;
+		theme->snapping_overlay_edge.bg_enabled = true;
+		theme->snapping_overlay_region.border_enabled = false;
+		theme->snapping_overlay_edge.border_enabled = false;
+	}
+
+	parse_hexstr("#8080b380", theme->snapping_overlay_region.bg_color);
+	parse_hexstr("#8080b380", theme->snapping_overlay_edge.bg_color);
+
+	/* inherit settings in post_processing() if not set elsewhere */
+	theme->snapping_overlay_region.border_width = INT_MIN;
+	theme->snapping_overlay_edge.border_width = INT_MIN;
+	memset(theme->snapping_overlay_region.border_color, 0,
+		sizeof(theme->snapping_overlay_region.border_color));
+	theme->snapping_overlay_region.border_color[0][0] = FLT_MIN;
+	memset(theme->snapping_overlay_edge.border_color, 0,
+		sizeof(theme->snapping_overlay_edge.border_color));
+	theme->snapping_overlay_edge.border_color[0][0] = FLT_MIN;
 }
 
 static void
@@ -627,6 +695,30 @@ entry(struct theme *theme, const char *key, const char *value)
 			theme->window_inactive_button_close_unpressed_image_color);
 	}
 
+	/* window drop-shadows */
+	if (match_glob(key, "window.active.shadow.size")) {
+		theme->window_active_shadow_size = atoi(value);
+		if (theme->window_active_shadow_size < 0) {
+			wlr_log(WLR_ERROR, "window.active.shadow.size cannot "
+				"be negative, clamping it to 0.");
+			theme->window_active_shadow_size = 0;
+		}
+	}
+	if (match_glob(key, "window.inactive.shadow.size")) {
+		theme->window_inactive_shadow_size = atoi(value);
+		if (theme->window_inactive_shadow_size < 0) {
+			wlr_log(WLR_ERROR, "window.inactive.shadow.size cannot "
+				"be negative, clamping it to 0.");
+			theme->window_inactive_shadow_size = 0;
+		}
+	}
+	if (match_glob(key, "window.active.shadow.color")) {
+		parse_hexstr(value, theme->window_active_shadow_color);
+	}
+	if (match_glob(key, "window.inactive.shadow.color")) {
+		parse_hexstr(value, theme->window_inactive_shadow_color);
+	}
+
 	if (match_glob(key, "menu.width.min")) {
 		theme->menu_min_width = atoi(value);
 	}
@@ -670,7 +762,12 @@ entry(struct theme *theme, const char *key, const char *value)
 		parse_hexstr(value, theme->osd_border_color);
 	}
 	if (match_glob(key, "osd.window-switcher.width")) {
-		theme->osd_window_switcher_width = atoi(value);
+		if (strrchr(value, '%')) {
+			theme->osd_window_switcher_width_is_percent = true;
+		} else {
+			theme->osd_window_switcher_width_is_percent = false;
+		}
+		theme->osd_window_switcher_width = MAX(atoi(value), 0);
 	}
 	if (match_glob(key, "osd.window-switcher.padding")) {
 		theme->osd_window_switcher_padding = atoi(value);
@@ -684,6 +781,12 @@ entry(struct theme *theme, const char *key, const char *value)
 	if (match_glob(key, "osd.window-switcher.item.active.border.width")) {
 		theme->osd_window_switcher_item_active_border_width = atoi(value);
 	}
+	if (match_glob(key, "osd.window-switcher.preview.border.width")) {
+		theme->osd_window_switcher_preview_border_width = atoi(value);
+	}
+	if (match_glob(key, "osd.window-switcher.preview.border.color")) {
+		parse_hexstrs(value, theme->osd_window_switcher_preview_border_color);
+	}
 	if (match_glob(key, "osd.workspace-switcher.boxes.width")) {
 		theme->osd_workspace_switcher_boxes_width = atoi(value);
 	}
@@ -693,6 +796,36 @@ entry(struct theme *theme, const char *key, const char *value)
 	if (match_glob(key, "osd.label.text.color")) {
 		parse_hexstr(value, theme->osd_label_text_color);
 	}
+	if (match_glob(key, "snapping.overlay.region.bg.enabled")) {
+		set_bool(value, &theme->snapping_overlay_region.bg_enabled);
+	}
+	if (match_glob(key, "snapping.overlay.edge.bg.enabled")) {
+		set_bool(value, &theme->snapping_overlay_edge.bg_enabled);
+	}
+	if (match_glob(key, "snapping.overlay.region.border.enabled")) {
+		set_bool(value, &theme->snapping_overlay_region.border_enabled);
+	}
+	if (match_glob(key, "snapping.overlay.edge.border.enabled")) {
+		set_bool(value, &theme->snapping_overlay_edge.border_enabled);
+	}
+	if (match_glob(key, "snapping.overlay.region.bg.color")) {
+		parse_hexstr(value, theme->snapping_overlay_region.bg_color);
+	}
+	if (match_glob(key, "snapping.overlay.edge.bg.color")) {
+		parse_hexstr(value, theme->snapping_overlay_edge.bg_color);
+	}
+	if (match_glob(key, "snapping.overlay.region.border.width")) {
+		theme->snapping_overlay_region.border_width = atoi(value);
+	}
+	if (match_glob(key, "snapping.overlay.edge.border.width")) {
+		theme->snapping_overlay_edge.border_width = atoi(value);
+	}
+	if (match_glob(key, "snapping.overlay.region.border.color")) {
+		parse_hexstrs(value, theme->snapping_overlay_region.border_color);
+	}
+	if (match_glob(key, "snapping.overlay.edge.border.color")) {
+		parse_hexstrs(value, theme->snapping_overlay_edge.border_color);
+	}
 }
 
 static void
@@ -884,8 +1017,8 @@ rounded_rect(struct rounded_corner_ctx *ctx)
 	 * We handle the edge-case where line-thickness > radius by merely
 	 * setting line-thickness = radius and in effect drawing a quadrant of a
 	 * circle. In this case the X and Y borders butt up against the arc and
-	 * overlap each other (as their line-thickessnes are greater than the
-	 * linethickness of the arc). As a result, there is no inner rounded
+	 * overlap each other (as their line-thicknesses are greater than the
+	 * line-thickness of the arc). As a result, there is no inner rounded
 	 * corners.
 	 *
 	 * So, in order to have inner rounded corners cornerRadius should be
@@ -950,6 +1083,193 @@ create_corners(struct theme *theme)
 	theme->corner_top_right_inactive_normal = rounded_rect(&ctx);
 }
 
+/*
+ * Draw the buffer used to render the edges of window drop-shadows. The buffer
+ * is 1 pixel tall and `visible_size` pixels wide and can be rotated and scaled for the
+ * different edges.  The buffer is drawn as would be found at the right-hand
+ * edge of a window. The gradient has a color of `start_color` at its left edge
+ * fading to clear at its right edge.
+ */
+static void
+shadow_edge_gradient(struct lab_data_buffer *buffer,
+		int visible_size, int total_size, float start_color[4])
+{
+	if (!buffer) {
+		/* This type of shadow is disabled, do nothing */
+		return;
+	}
+
+	assert(buffer->format == DRM_FORMAT_ARGB8888);
+	uint8_t *pixels = buffer->data;
+
+	/* Inset portion which is obscured */
+	int inset = total_size - visible_size;
+
+	/* Standard deviation normalised against the shadow width, squared */
+	double variance = 0.3 * 0.3;
+
+	for (int x = 0; x < visible_size; x++) {
+		/*
+		 * x normalised against total shadow width. We add on inset here
+		 * because we don't bother drawing inset for the edge shadow
+		 * buffers but still need the pattern to line up with the corner
+		 * shadow buffers which do have inset drawn.
+		 */
+		double xn = (double)(x + inset) / (double)total_size;
+
+		/* Gausian dropoff */
+		double alpha = exp(-(xn * xn) / variance);
+
+		/* RGBA values are all pre-multiplied */
+		pixels[4 * x] = start_color[2] * alpha * 255;
+		pixels[4 * x + 1] = start_color[1] * alpha * 255;
+		pixels[4 * x + 2] = start_color[0] * alpha * 255;
+		pixels[4 * x + 3] = start_color[3] * alpha * 255;
+	}
+}
+
+/*
+ * Draw the buffer used to render the corners of window drop-shadows.  The
+ * shadow looks better if the buffer is inset behind the window, so the buffer
+ * is square with a size of radius+inset.  The buffer is drawn for the
+ * bottom-right corner but can be rotated for other corners.  The gradient fades
+ * from `start_color` at the top-left to clear at the opposite edge.
+ *
+ * If the window is translucent we don't want the shadow to be visible through
+ * it.  For the bottom corners of the window this is easy, we just erase the
+ * square of the buffer which will be behind the window.  For the top it's a
+ * little more complicated because the titlebar can have rounded corners.
+ * However, the titlebar itself is always opaque so we only have to erase the
+ * L-shaped area of the buffer which can appear behind the non-titlebar part of
+ * the window.
+ */
+static void
+shadow_corner_gradient(struct lab_data_buffer *buffer, int visible_size,
+	int total_size, int titlebar_height, float start_color[4])
+{
+	if (!buffer) {
+		/* This type of shadow is disabled, do nothing */
+		return;
+	}
+
+	assert(buffer->format == DRM_FORMAT_ARGB8888);
+	uint8_t *pixels = buffer->data;
+
+	/* Standard deviation normalised against the shadow width, squared */
+	double variance = 0.3 * 0.3;
+
+	int inset = total_size - visible_size;
+
+	for (int y = 0; y < total_size; y++) {
+		uint8_t *pixel_row = &pixels[y * buffer->stride];
+		for (int x = 0; x < total_size; x++) {
+			/* x and y normalised against total shadow width */
+			double x_norm = (double)(x) / (double)total_size;
+			double y_norm = (double)(y) / (double)total_size;
+			/*
+			 * For Gaussian drop-off in 2d you can just calculate
+			 * the outer product of the horizontal and vertical
+			 * profiles.
+			 */
+			double gauss_x = exp(-(x_norm * x_norm) / variance);
+			double gauss_y = exp(-(y_norm * y_norm) / variance);
+			double alpha = gauss_x * gauss_y;
+
+			/*
+			 * Erase the L-shaped region which could be visible
+			 * through a transparent window but not obscured by the
+			 * titlebar. If inset is smaller than the titlebar
+			 * height then there's nothing to do, this is handled by
+			 * (inset - titlebar_height) being negative.
+			 */
+			bool in1 = x < inset && y < inset - titlebar_height;
+			bool in2 = x < inset - titlebar_height && y < inset;
+			if (in1 || in2) {
+				alpha = 0.0;
+			}
+
+			/* RGBA values are all pre-multiplied */
+			pixel_row[4 * x] = start_color[2] * alpha * 255;
+			pixel_row[4 * x + 1] = start_color[1] * alpha * 255;
+			pixel_row[4 * x + 2] = start_color[0] * alpha * 255;
+			pixel_row[4 * x + 3] = start_color[3] * alpha * 255;
+		}
+	}
+}
+
+static void
+create_shadows(struct theme *theme)
+{
+	/* Size of shadow visible extending beyond the window */
+	int visible_active_size = theme->window_active_shadow_size;
+	int visible_inactive_size = theme->window_inactive_shadow_size;
+	/* How far inside the window the shadow inset begins */
+	int inset_active = (double)visible_active_size * SSD_SHADOW_INSET;
+	int inset_inactive = (double)visible_inactive_size * SSD_SHADOW_INSET;
+	/* Total width including visible and obscured portion */
+	int total_active_size = visible_active_size + inset_active;
+	int total_inactive_size = visible_inactive_size + inset_inactive;
+
+	/*
+	 * Edge shadows don't need to be inset so the buffers are sized just for
+	 * the visible width.  Corners are inset so the buffers are larger for
+	 * this.
+	 */
+	if (visible_active_size > 0) {
+		theme->shadow_edge_active = buffer_create_cairo(
+			visible_active_size, 1, 1.0, true);
+		theme->shadow_corner_top_active = buffer_create_cairo(
+			total_active_size, total_active_size, 1.0, true);
+		theme->shadow_corner_bottom_active = buffer_create_cairo(
+			total_active_size, total_active_size, 1.0, true);
+		if (!theme->shadow_corner_top_active
+				|| !theme->shadow_corner_bottom_active
+				|| !theme->shadow_edge_active) {
+			wlr_log(WLR_ERROR, "Failed to allocate shadow buffer");
+			return;
+		}
+	}
+	if (visible_inactive_size > 0) {
+		theme->shadow_edge_inactive = buffer_create_cairo(
+			visible_inactive_size, 1, 1.0, true);
+		theme->shadow_corner_top_inactive = buffer_create_cairo(
+			total_inactive_size, total_inactive_size, 1.0, true);
+		theme->shadow_corner_bottom_inactive = buffer_create_cairo(
+			total_inactive_size, total_inactive_size, 1.0, true);
+		if (!theme->shadow_corner_top_inactive
+				|| !theme->shadow_corner_bottom_inactive
+				|| !theme->shadow_edge_inactive) {
+			wlr_log(WLR_ERROR, "Failed to allocate shadow buffer");
+			return;
+		}
+	}
+
+	shadow_edge_gradient(theme->shadow_edge_active, visible_active_size,
+		total_active_size, theme->window_active_shadow_color);
+	shadow_edge_gradient(theme->shadow_edge_inactive, visible_inactive_size,
+		total_inactive_size, theme->window_inactive_shadow_color);
+	shadow_corner_gradient(theme->shadow_corner_top_active,
+		visible_active_size, total_active_size,
+		theme->title_height, theme->window_active_shadow_color);
+	shadow_corner_gradient(theme->shadow_corner_bottom_active,
+		visible_active_size, total_active_size, 0,
+		theme->window_active_shadow_color);
+	shadow_corner_gradient(theme->shadow_corner_top_inactive,
+		visible_inactive_size, total_inactive_size,
+		theme->title_height, theme->window_inactive_shadow_color);
+	shadow_corner_gradient(theme->shadow_corner_bottom_inactive,
+		visible_inactive_size, total_inactive_size, 0,
+		theme->window_inactive_shadow_color);
+}
+
+static void
+fill_colors_with_osd_theme(struct theme *theme, float colors[3][4])
+{
+	memcpy(colors[0], theme->osd_bg_color, sizeof(colors[0]));
+	memcpy(colors[1], theme->osd_label_text_color, sizeof(colors[1]));
+	memcpy(colors[2], theme->osd_bg_color, sizeof(colors[2]));
+}
+
 static void
 post_processing(struct theme *theme)
 {
@@ -1007,16 +1327,45 @@ post_processing(struct theme *theme)
 	if (theme->osd_workspace_switcher_boxes_height == 0) {
 		theme->osd_workspace_switcher_boxes_width = 0;
 	}
+	if (theme->osd_window_switcher_width_is_percent) {
+		theme->osd_window_switcher_width =
+			MIN(theme->osd_window_switcher_width, 100);
+	}
+	if (theme->osd_window_switcher_preview_border_width == INT_MIN) {
+		theme->osd_window_switcher_preview_border_width =
+			theme->osd_border_width;
+	}
+	if (theme->osd_window_switcher_preview_border_color[0][0] == FLT_MIN) {
+		fill_colors_with_osd_theme(theme,
+			theme->osd_window_switcher_preview_border_color);
+	}
+
+	if (theme->snapping_overlay_region.border_width == INT_MIN) {
+		theme->snapping_overlay_region.border_width =
+			theme->osd_border_width;
+	}
+	if (theme->snapping_overlay_edge.border_width == INT_MIN) {
+		theme->snapping_overlay_edge.border_width =
+			theme->osd_border_width;
+	}
+	if (theme->snapping_overlay_region.border_color[0][0] == FLT_MIN) {
+		fill_colors_with_osd_theme(theme,
+			theme->snapping_overlay_region.border_color);
+	}
+	if (theme->snapping_overlay_edge.border_color[0][0] == FLT_MIN) {
+		fill_colors_with_osd_theme(theme,
+			theme->snapping_overlay_edge.border_color);
+	}
 }
 
 void
-theme_init(struct theme *theme, const char *theme_name)
+theme_init(struct theme *theme, struct server *server, const char *theme_name)
 {
 	/*
 	 * Set some default values. This is particularly important on
 	 * reconfigure as not all themes set all options
 	 */
-	theme_builtin(theme);
+	theme_builtin(theme, server);
 
 	/* Read <data-dir>/share/themes/$theme_name/openbox-3/themerc */
 	struct wl_list paths;
@@ -1032,6 +1381,7 @@ theme_init(struct theme *theme, const char *theme_name)
 	post_processing(theme);
 	create_corners(theme);
 	load_buttons(theme);
+	create_shadows(theme);
 }
 
 void
@@ -1041,4 +1391,10 @@ theme_finish(struct theme *theme)
 	zdrop(&theme->corner_top_left_inactive_normal);
 	zdrop(&theme->corner_top_right_active_normal);
 	zdrop(&theme->corner_top_right_inactive_normal);
+	zdrop(&theme->shadow_corner_top_active);
+	zdrop(&theme->shadow_corner_bottom_active);
+	zdrop(&theme->shadow_edge_active);
+	zdrop(&theme->shadow_corner_top_inactive);
+	zdrop(&theme->shadow_corner_bottom_inactive);
+	zdrop(&theme->shadow_edge_inactive);
 }
