@@ -2,6 +2,7 @@
 #include <assert.h>
 #include <stdio.h>
 #include <strings.h>
+#include <wlr/types/wlr_output_layout.h>
 #include "common/macros.h"
 #include "common/match.h"
 #include "common/mem.h"
@@ -9,9 +10,11 @@
 #include "input/keyboard.h"
 #include "labwc.h"
 #include "menu/menu.h"
+#include "osd.h"
 #include "placement.h"
 #include "regions.h"
 #include "resize_indicator.h"
+#include "snap-constraints.h"
 #include "snap.h"
 #include "ssd.h"
 #include "view.h"
@@ -107,6 +110,11 @@ matches_criteria(struct view *view, enum lab_view_criteria criteria)
 			return false;
 		}
 	}
+	if (criteria & LAB_VIEW_CRITERIA_ROOT_TOPLEVEL) {
+		if (view != view_get_root(view)) {
+			return false;
+		}
+	}
 	if (criteria & LAB_VIEW_CRITERIA_NO_ALWAYS_ON_TOP) {
 		if (view_is_always_on_top(view)) {
 			return false;
@@ -136,6 +144,48 @@ view_next(struct wl_list *head, struct view *view, enum lab_view_criteria criter
 	return NULL;
 }
 
+struct view *
+view_next_no_head_stop(struct wl_list *head, struct view *from,
+		enum lab_view_criteria criteria)
+{
+	assert(head);
+
+	struct wl_list *elm = from ? &from->link : head;
+
+	struct wl_list *end = elm;
+	for (elm = elm->next; elm != end; elm = elm->next) {
+		if (elm == head) {
+			continue;
+		}
+		struct view *view = wl_container_of(elm, view, link);
+		if (matches_criteria(view, criteria)) {
+			return view;
+		}
+	}
+	return from;
+}
+
+struct view *
+view_prev_no_head_stop(struct wl_list *head, struct view *from,
+		enum lab_view_criteria criteria)
+{
+	assert(head);
+
+	struct wl_list *elm = from ? &from->link : head;
+
+	struct wl_list *end = elm;
+	for (elm = elm->prev; elm != end; elm = elm->prev) {
+		if (elm == head) {
+			continue;
+		}
+		struct view *view = wl_container_of(elm, view, link);
+		if (matches_criteria(view, criteria)) {
+			return view;
+		}
+	}
+	return from;
+}
+
 void
 view_array_append(struct server *server, struct wl_array *views,
 		enum lab_view_criteria criteria)
@@ -162,28 +212,26 @@ view_wants_focus(struct view *view)
 }
 
 bool
-view_is_focusable_from(struct view *view, struct wlr_surface *prev)
+view_contains_window_type(struct view *view, enum window_type window_type)
+{
+	assert(view);
+	if (view->impl->contains_window_type) {
+		return view->impl->contains_window_type(view, window_type);
+	}
+	return false;
+}
+
+bool
+view_is_focusable(struct view *view)
 {
 	assert(view);
 	if (!view->surface) {
 		return false;
 	}
-	if (!view->mapped && !view->minimized) {
+	if (view_wants_focus(view) != VIEW_WANTS_FOCUS_ALWAYS) {
 		return false;
 	}
-	enum view_wants_focus wants_focus = view_wants_focus(view);
-	/*
-	 * Consider "offer focus" (Globally Active) views as focusable
-	 * only if another surface from the same application already had
-	 * focus. The goal is to allow focusing a parent window when a
-	 * dialog/popup is closed, but still avoid focusing standalone
-	 * panels/toolbars/notifications. Note that we are basically
-	 * guessing whether Globally Active views want focus, and will
-	 * probably be wrong some of the time.
-	 */
-	return (wants_focus == VIEW_WANTS_FOCUS_ALWAYS
-		|| (wants_focus == VIEW_WANTS_FOCUS_OFFER
-			&& prev && view_is_related(view, prev)));
+	return (view->mapped || view->minimized);
 }
 
 /**
@@ -600,6 +648,11 @@ view_minimize(struct view *view, bool minimized)
 	struct view *root = view_get_root(view);
 	_minimize(root, minimized);
 	minimize_sub_views(root, minimized);
+
+	/* Enable top-layer when full-screen views are minimized */
+	if (view->fullscreen && view->output) {
+		desktop_update_top_layer_visiblity(view->server);
+	}
 }
 
 bool
@@ -1519,8 +1572,44 @@ view_on_output_destroy(struct view *view)
 	view->output = NULL;
 }
 
+static enum wlr_direction
+opposite_direction(enum wlr_direction direction)
+{
+	switch (direction) {
+	case WLR_DIRECTION_RIGHT:
+		return WLR_DIRECTION_LEFT;
+	case WLR_DIRECTION_LEFT:
+		return WLR_DIRECTION_RIGHT;
+	case WLR_DIRECTION_DOWN:
+		return WLR_DIRECTION_UP;
+	case WLR_DIRECTION_UP:
+		return WLR_DIRECTION_DOWN;
+	default:
+		return 0;
+	}
+}
+
+static enum wlr_direction
+get_wlr_direction(enum view_edge edge)
+{
+	switch (edge) {
+	case VIEW_EDGE_LEFT:
+		return WLR_DIRECTION_LEFT;
+	case VIEW_EDGE_RIGHT:
+		return WLR_DIRECTION_RIGHT;
+	case VIEW_EDGE_UP:
+		return WLR_DIRECTION_UP;
+	case VIEW_EDGE_DOWN:
+		return WLR_DIRECTION_DOWN;
+	case VIEW_EDGE_CENTER:
+	case VIEW_EDGE_INVALID:
+	default:
+		return 0;
+	}
+}
+
 struct output *
-view_get_adjacent_output(struct view *view, enum view_edge edge)
+view_get_adjacent_output(struct view *view, enum view_edge edge, bool wrap)
 {
 	assert(view);
 	struct output *output = view->output;
@@ -1530,32 +1619,31 @@ view_get_adjacent_output(struct view *view, enum view_edge edge)
 		return NULL;
 	}
 
+	struct wlr_box box = output_usable_area_in_layout_coords(output);
+	int lx = box.x + box.width / 2;
+	int ly = box.y + box.height / 2;
+
 	/* Determine any adjacent output in the appropriate direction */
 	struct wlr_output *new_output = NULL;
 	struct wlr_output *current_output = output->wlr_output;
 	struct wlr_output_layout *layout = view->server->output_layout;
-	switch (edge) {
-	case VIEW_EDGE_LEFT:
-		new_output = wlr_output_layout_adjacent_output(
-			layout, WLR_DIRECTION_LEFT, current_output, 1, 0);
-		break;
-	case VIEW_EDGE_RIGHT:
-		new_output = wlr_output_layout_adjacent_output(
-			layout, WLR_DIRECTION_RIGHT, current_output, 1, 0);
-		break;
-	case VIEW_EDGE_UP:
-		new_output = wlr_output_layout_adjacent_output(
-			layout, WLR_DIRECTION_UP, current_output, 0, 1);
-		break;
-	case VIEW_EDGE_DOWN:
-		new_output = wlr_output_layout_adjacent_output(
-			layout, WLR_DIRECTION_DOWN, current_output, 0, 1);
-		break;
-	default:
-		break;
+	enum wlr_direction direction = get_wlr_direction(edge);
+	new_output = wlr_output_layout_adjacent_output(layout, direction,
+		current_output, lx, ly);
+
+	/*
+	 * Optionally wrap around from top-to-bottom or left-to-right, and vice
+	 * versa.
+	 */
+	if (wrap && !new_output) {
+		new_output = wlr_output_layout_farthest_output(layout,
+			opposite_direction(direction), current_output, lx, ly);
 	}
 
-	/* When "adjacent" output is the same as the original, there is no adjacent */
+	/*
+	 * When "adjacent" output is the same as the original, there is no
+	 * adjacent
+	 */
 	if (!new_output || new_output == current_output) {
 		return NULL;
 	}
@@ -1629,7 +1717,8 @@ view_move_to_edge(struct view *view, enum view_edge direction, bool snap_to_wind
 	}
 
 	/* Otherwise, move to edge of next adjacent display, if possible */
-	struct output *output = view_get_adjacent_output(view, direction);
+	struct output *output =
+		view_get_adjacent_output(view, direction, /* wrap */ false);
 	if (!output) {
 		return;
 	}
@@ -1788,7 +1877,7 @@ view_snap_to_edge(struct view *view, enum view_edge edge,
 
 	if (across_outputs && view->tiled == edge && view->maximized == VIEW_AXIS_NONE) {
 		/* We are already tiled for this edge; try to switch outputs */
-		output = view_get_adjacent_output(view, edge);
+		output = view_get_adjacent_output(view, edge, /* wrap */ false);
 
 		if (!output) {
 			/*
@@ -1982,17 +2071,6 @@ view_append_children(struct view *view, struct wl_array *children)
 	}
 }
 
-bool
-view_is_related(struct view *view, struct wlr_surface *surface)
-{
-	assert(view);
-	assert(surface);
-	if (view->impl->is_related) {
-		return view->impl->is_related(view, surface);
-	}
-	return false;
-}
-
 bool
 view_has_strut_partial(struct view *view)
 {
@@ -2134,6 +2212,10 @@ view_set_shade(struct view *view, bool shaded)
 	view->shaded = shaded;
 	ssd_enable_shade(view->ssd, view->shaded);
 	wlr_scene_node_set_enabled(view->scene_node, !view->shaded);
+
+	if (view->impl->shade) {
+		view->impl->shade(view, shaded);
+	}
 }
 
 void
@@ -2142,6 +2224,8 @@ view_destroy(struct view *view)
 	assert(view);
 	struct server *server = view->server;
 
+	snap_constraints_invalidate(view);
+
 	if (view->mappable.connected) {
 		mappable_disconnect(&view->mappable);
 	}
@@ -2162,7 +2246,7 @@ view_destroy(struct view *view)
 		/* Application got killed while moving around */
 		server->input_mode = LAB_INPUT_STATE_PASSTHROUGH;
 		server->grabbed_view = NULL;
-		regions_hide_overlay(&server->seat);
+		overlay_hide(&server->seat);
 	}
 
 	if (server->active_view == view) {
