@@ -8,6 +8,7 @@
 #include "decorations.h"
 #include "labwc.h"
 #include "node.h"
+#include "snap-constraints.h"
 #include "view.h"
 #include "view-impl-common.h"
 #include "window-rules.h"
@@ -42,6 +43,28 @@ xdg_toplevel_from_view(struct view *view)
 	return xdg_surface->toplevel;
 }
 
+static bool
+xdg_toplevel_view_contains_window_type(struct view *view, int32_t window_type)
+{
+	assert(view);
+
+	struct wlr_xdg_toplevel *toplevel = xdg_toplevel_from_view(view);
+	struct wlr_xdg_toplevel_state *state = &toplevel->current;
+	bool is_dialog = (state->min_width != 0 && state->min_height != 0
+		&& (state->min_width == state->max_width
+		|| state->min_height == state->max_height))
+		|| toplevel->parent;
+
+	switch (window_type) {
+	case NET_WM_WINDOW_TYPE_NORMAL:
+		return !is_dialog;
+	case NET_WM_WINDOW_TYPE_DIALOG:
+		return is_dialog;
+	default:
+		return false;
+	}
+}
+
 static void
 handle_new_popup(struct wl_listener *listener, void *data)
 {
@@ -141,6 +164,7 @@ handle_commit(struct wl_listener *listener, void *data)
 		 * actual view.
 		 */
 		if (!view->pending_configure_serial) {
+			snap_constraints_update(view);
 			view->pending = view->current;
 
 			/*
@@ -148,7 +172,7 @@ handle_commit(struct wl_listener *listener, void *data)
 			 * wlr_xdg_toplevel_set_size and will send the retained
 			 * values with every subsequent configure request. If a
 			 * client has resized itself in the meantime, a
-			 * configure request that sends the now-outated size
+			 * configure request that sends the now-outdated size
 			 * may prompt the client to resize itself unexpectedly.
 			 *
 			 * Calling wlr_xdg_toplevel_set_size to update the
@@ -188,6 +212,7 @@ handle_configure_timeout(void *data)
 		view->current.width, view->current.height);
 
 	/* Re-sync pending view with current state */
+	snap_constraints_update(view);
 	view->pending = view->current;
 
 	return 0; /* ignored per wl_event_loop docs */
@@ -236,13 +261,14 @@ handle_request_move(struct wl_listener *listener, void *data)
 	 * This event is raised when a client would like to begin an interactive
 	 * move, typically because the user clicked on their client-side
 	 * decorations. Note that a more sophisticated compositor should check
-	 * the provied serial against a list of button press serials sent to
-	 * this
-	 * client, to prevent the client from requesting this whenever they
+	 * the provided serial against a list of button press serials sent to
+	 * this client, to prevent the client from requesting this whenever they
 	 * want.
 	 */
 	struct view *view = wl_container_of(listener, view, request_move);
-	interactive_begin(view, LAB_INPUT_STATE_MOVE, 0);
+	if (view == view->server->seat.pressed.view) {
+		interactive_begin(view, LAB_INPUT_STATE_MOVE, 0);
+	}
 }
 
 static void
@@ -252,14 +278,15 @@ handle_request_resize(struct wl_listener *listener, void *data)
 	 * This event is raised when a client would like to begin an interactive
 	 * resize, typically because the user clicked on their client-side
 	 * decorations. Note that a more sophisticated compositor should check
-	 * the provied serial against a list of button press serials sent to
-	 * this
-	 * client, to prevent the client from requesting this whenever they
+	 * the provided serial against a list of button press serials sent to
+	 * this client, to prevent the client from requesting this whenever they
 	 * want.
 	 */
 	struct wlr_xdg_toplevel_resize_event *event = data;
 	struct view *view = wl_container_of(listener, view, request_resize);
-	interactive_begin(view, LAB_INPUT_STATE_RESIZE, event->edges);
+	if (view == view->server->seat.pressed.view) {
+		interactive_begin(view, LAB_INPUT_STATE_RESIZE, event->edges);
+	}
 }
 
 static void
@@ -632,6 +659,20 @@ xdg_toplevel_view_unmap(struct view *view, bool client_request)
 	}
 }
 
+static pid_t
+xdg_view_get_pid(struct view *view)
+{
+	assert(view);
+	pid_t pid = -1;
+
+	if (view->surface && view->surface->resource
+			&& view->surface->resource->client) {
+		struct wl_client *client = view->surface->resource->client;
+		wl_client_get_credentials(client, &pid, NULL, NULL);
+	}
+	return pid;
+}
+
 static const struct view_impl xdg_toplevel_view_impl = {
 	.configure = xdg_toplevel_view_configure,
 	.close = xdg_toplevel_view_close,
@@ -647,6 +688,8 @@ static const struct view_impl xdg_toplevel_view_impl = {
 	.move_to_back = view_impl_move_to_back,
 	.get_root = xdg_toplevel_view_get_root,
 	.append_children = xdg_toplevel_view_append_children,
+	.contains_window_type = xdg_toplevel_view_contains_window_type,
+	.get_pid = xdg_view_get_pid,
 };
 
 static void
@@ -682,6 +725,11 @@ xdg_activation_handle_request(struct wl_listener *listener, void *data)
 		return;
 	}
 
+	if (view->server->osd_state.cycle_view) {
+		wlr_log(WLR_INFO, "Preventing focus request while in window switcher");
+		return;
+	}
+
 	wlr_log(WLR_DEBUG, "Activating surface");
 	desktop_focus_view(view, /*raise*/ true);
 }
@@ -771,7 +819,7 @@ xdg_surface_new(struct wl_listener *listener, void *data)
 	 */
 	kde_server_decoration_set_view(view, xdg_surface->surface);
 
-	/* In support of xdg popups */
+	/* In support of xdg popups and IME popup */
 	xdg_surface->surface->data = tree;
 
 	view_connect_map(view, xdg_surface->surface);
